/*****************************************************************************/
/* This GAUSS6.0 Procedure is for Pair-Wise test of M.H.Pesaran (2007),     **/
/*   A Pair-wise Approach to Testing for Output and Growth Convergence,     **/
/*   Journal of Econometrics 127,312-355                                    **/

/**         {fracmat,frac_kpss}=pairwise(x_mat,maxp,case);                  **/
/** <IMPUTS>                                                                **/
/** x_mat: panel data matrix ordered TxN,  where T is maximum number of     **/
/**        time series dimension, N is number of cross section dimension    **/
/** maxp: maximum lag-order for Akaike Information Criteria (AIC) and       **/
/**       Schwartz Basian Information Criterion (SBC) to choose augmentation**/
/**       lag-order in each ADF type regression (the choice of lag-order    **/
/**       in ADF regression is used for that in ADF-GLS and ADF-WS          **/
/** case: Deterministics in ADF type regression.                            **/
/**         case=2: an intercept without linear time trend                  **/
/**         case=3: an intercept with a linear time trend                   **/
/** <OUTPUT>								    **/
/** output is printed in the command window.                                **/
/** fracmat: 6x2 matrix of fraction of rejections of the null hypothesis    **/
/**         of nonstationarity (ADFtype) over N(N-1)/2 pairs:		        **/

/**		ADF-AIC5% 	    ADF-SBC5%
		ADF-GLS-AIC5%	ADF-GLS-SBC5%
		ADF-WS-AIC5%	ADF-WS-AIC5%
		ADF-AIC10% 	    ADF-SBC10%
		ADF-GLS-AIC10%	ADF-GLS-SBC10%
		ADF-WS-AIC10%	ADF-WS-AIC10%	                         
							**/
/** frac_kpss: 2x1 vector of fraction of rejections of the null hypothesis  **/
/**         of stationarity (ADFtype) by (KPSS) test over N(N-1)/2 pairs:   **/
                                               
/**		kpss5%
		kpss10%					**/
                    
/** Unbalanced panel data is allowed but only at the begining or at the end **/
/** of the data period (leave the excel cell blank)                         **/
/**  					                    Takashi Yamagata  11 July 2008  **/
/*****************************************************************************/

proc(2)=pairwise(x_mat,maxp,case);
local dfcv,dfglscv,dfwscv,crtkpss,n,tvec,tijvec,tadf_vec,tadf_gls_vec,kpss_vec,pvec_aic,pvec_sbc;
local dfcv10,dfglscv10,dfwscv10,crtkpss10,casename;
local i,x_i,j,x_j,x_ij,tadf,aic_adf,sbc_adf,non,non2,tadf_aic,tadf_sbc;
local x_ij_gls,tadf_gls,aic_adf_gls,sbc_adf_gls,tadf_gls_aic,tadf_gls_sbc;
local x_ij_ws,c_p,tadf_ws,aic_adf_ws,sbc_adf_ws,tadfws,aicws,sbcws;
local tadf_ws_aic,tadf_ws_sbc,kpss_stat1,tadf_ws_vec;
local frac_tadf,frac_tadf_gls,frac_tadf_ws,frac_kpss,fracmat;
local frac_tadf10,frac_tadf_gls10,frac_tadf_ws10,frac_kpss10,fracmat10;

/* Set critical values */
if case==2;/** constant, no trend **/
casename="Case II: An Intercept Only";
dfcv=-2.86;/* 5% critical value of ADF test **/
dfglscv=-1.94;/* 5% critical value of ADF-GLS test **/
dfwscv=-2.50;/* 5% critical value of ADF-WS test **/
crtkpss=0.463;/* 5% critical value of KPSS test **/

dfcv10=-2.57;/* 10% critical value of ADF test **/
dfglscv10=-1.63;/* 10% critical value of ADF-GLS test **/
dfwscv10=-2.22;/* 10% critical value of ADF-WS test (computed based on T=2500,10000 reps) **/
crtkpss10=0.347;/* 10% critical value of KPSS test **/


elseif case==3;/** constant and trend **/
casename="Case III: An Intercept and a Linear Trend";
dfcv=-3.41;/* 5% critical value of ADF test **/
dfglscv=-2.86;/* 5% critical value of ADF-GLS test **/
dfwscv=-3.19;/* 5% critical value of ADF-WS test **/
crtkpss=0.146;/* 5% critical value of KPSS test **/

dfcv10=-3.12;/* 10% critical value of ADF test **/
dfglscv10=-2.57;/* 10% critical value of ADF-GLS test **/
dfwscv10=-2.90;/* 10% critical value of ADF-WS test (computed based on T=2500,10000 reps) **/
crtkpss10=0.119;/* 10% critical value of KPSS test **/




endif;


        n=cols(x_mat);
tvec={};
tijvec={};
    /* (2): Pair-Wise Test  */

tadf_vec={};
tadf_gls_vec={};
tadf_ws_vec={};
kpss_vec={};
pvec_aic={};
pvec_sbc={};

i=1;
do while i<=n-1;
x_i=x_mat[.,i];
tvec=tvec|rows(delif(x_i,x_i.==_dxmiss));
	j=i+1;
	do while j<=n;
		x_j=x_mat[.,j];
		x_ij=x_i-x_j; 
        x_ij=delif(x_ij,x_ij.==_dxmiss);/*** delete missing observations ****/
            tijvec=tijvec|rows(x_ij);
/** you need to obtain correct T_ijx1 vector x_ij **/
                /** ADF(p) **/
                {tadf,aic_adf,sbc_adf,non,non2}=adf(x_ij,maxp,case);
                tadf_aic=tadf[minindc(aic_adf)];
                tadf_sbc=tadf[minindc(sbc_adf)];
                /** ADF-GLS **/
                x_ij_gls = glsdetrend(x_ij,case-2);
                {tadf_gls,aic_adf_gls,sbc_adf_gls,non,non2}=adf(x_ij_gls,maxp,1);
                tadf_gls_aic=tadf_gls[minindc(aic_adf)];
                tadf_gls_sbc=tadf_gls[minindc(sbc_adf)];

                /** ADF-WS **/
                x_ij_ws = detrend(x_ij,case-2);
                c_p=0;tadf_ws=zeros(maxp+1,1);aic_adf_ws=zeros(maxp+1,1);sbc_adf_ws=zeros(maxp+1,1);
                    do while c_p<=maxp;    
                        {tadfws,aicws,sbcws}= wsp(x_ij_ws,case-2,c_p);
                        tadf_ws[c_p+1]=tadfws;
                        aic_adf_ws[c_p+1]=aicws;sbc_adf_ws[c_p+1]=sbcws;
                        c_p=c_p+1;
                    endo;
                pvec_aic=pvec_aic|minindc(aic_adf);
                pvec_sbc=pvec_sbc|minindc(sbc_adf);
                tadf_ws_aic=tadf_ws[minindc(aic_adf)];
                tadf_ws_sbc=tadf_ws[minindc(sbc_adf)];

                /** KPSS **/
		/** the last digit "1" is the choice of band width (T^(1/3))*0.75*/
                kpss_stat1=kpss(x_ij,case,1);


                /*** Reject=1, accept=0 ***/
                    tadf_vec=tadf_vec|(tadf_aic~tadf_sbc);
                    tadf_gls_vec=tadf_gls_vec|(tadf_gls_aic~tadf_gls_sbc);
                    tadf_ws_vec=tadf_ws_vec|(tadf_ws_aic~tadf_ws_sbc);
                    kpss_vec=kpss_vec|((kpss_stat1));

	j=j+1;
	endo;
i=i+1;
endo;
/*5%test*/
frac_tadf=meanc(tadf_vec.<dfcv);
frac_tadf_gls=meanc(tadf_gls_vec.<dfglscv);
frac_tadf_ws=meanc(tadf_ws_vec.<dfwscv);
frac_kpss=meanc(kpss_vec.>crtkpss);
/*10% test*/
frac_tadf10=meanc(tadf_vec.<dfcv10);
frac_tadf_gls10=meanc(tadf_gls_vec.<dfglscv10);
frac_tadf_ws10=meanc(tadf_ws_vec.<dfwscv10);
frac_kpss10=meanc(kpss_vec.>crtkpss10);

fracmat=(frac_tadf~frac_tadf_gls~frac_tadf_ws)';
fracmat10=(frac_tadf10~frac_tadf_gls10~frac_tadf_ws10)';

/** etcetra */
"+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++";
"<<Pair-Wise test of M.H.Pesaran (2007), A Pair-wise Approach to Testing 
   for Output and Growth Convergence, Journal of Econometrics 127,312-355>>";
"N=";;n;;", N(N-1)/2=";;n*(n-1)/2;
"Max(Ti)=";;maxc(tvec);;", Min(Ti)=";;minc(tvec);
"Max(Tij)=";;maxc(tijvec);;", Min(Tij)=";;minc(tijvec);
"Average Ti:";;meanc(tvec);
"Average Tij:";;meanc(tijvec);
"Maximum Lag-Order was";;maxp;
"   Average Lag-Order by AIC:";;meanc(pvec_aic);
"   Average Lag-Order by SBC:";;meanc(pvec_sbc);
"----------------------------------";
"[Fraction of Rejections, based on 5% nominal level tests]";
$casename;
$(" "~"AIC"~"SBC")|(
("ADF"|"ADF-GLS"|"ADF-WS")
~ftocv(fracmat,1,5));
"";
$"KPSS"~ftocv(frac_kpss,1,5);
"";
"* Each test is conducted at the 5% significance level.";
"* In this code the critical values are NOT depending on T and lag-order p.";
"* Bandwidth for KPSS test is round((T^(1/3))*0.75)";
"----------------------------------";
"[Fraction of Rejections, based on 10% nominal level tests]";
$casename;
$(" "~"AIC"~"SBC")|(
("ADF"|"ADF-GLS"|"ADF-WS")
~ftocv(fracmat10,1,5));
"";
$"KPSS"~ftocv(frac_kpss10,1,5);
"";
"* Each test is conducted at the 10% significance level.";
"* In this code the critical values are NOT depending on T and lag-order p.";
"* Bandwidth for KPSS test is round((T^(1/3))*0.75)";
"+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++";

retp(fracmat|fracmat10,frac_kpss|frac_kpss10);
endp;


/*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@*/
/*@@@@@@@@@@@@@@@@@@@@@@@@@@ PROCEDURES TO BE USED  @@@@@@@@@@@@@@@@@@@@@@@*/
/*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@*/
/**************************************************************************/



library pgraph;
/*
** GLSDETREND
**
** Format:  {yd} = glsdetrend(y,d);
**
** Input:   y     -- time series variable
**
**          d     -- order of the time-polynomial to include in the
**                   ADF regression. Set d = 0 for a conctant. Set
**                    d = 1 for a constant plus trend. c-bar used for
**                    detrending is -7 for d=0 and -13.5 for d=1 taken
**                   from Elliot et al. (1996), Econometrica, 64, p.813-836.
**
** output:  yd       --  gls detrended series
**
**
*/
proc glsdetrend(y,d);
local n, alp, c, ct, z, lz, zalp, ly, yalp, b, yd;

n=rows(y);

c=-7;
ct=-13.5;


   if d==0;

         alp = 1 + c/n;

         z = ones(n,1);


         lz = lagn(z,1); lz[1] = 0;

   elseif d==1;

         alp = 1 + ct/n;

         z = ones(n,1)~seqa(1,1,n);

         lz = lagn(z,1); lz[1,.] = zeros(1,2);

   endif;

         zalp = z - alp*lz;


         ly = lagn(y,1); ly[1] = 0;

         yalp = y - alp*ly;

         b = inv(zalp'*zalp)*zalp'*yalp;

         yd = y - z*b;

retp(yd);
endp;





/*
** WS  
**
** Format:  {ws} = wsp(x,d,p);
**
** Input:   x     -- time series variable 
**
**	    d     -- type of detrending prior to implementing the weighted symmetric procedure 
**                   d=0: if constant is removed
**                   d=1: if constant and trend is removed
**
**          p     -- number of lagged changes of x to include in the
**                   fitted regression.
**                    
** output:  wsstat    -- weighted symmetric statistic
**
**NOTE: No deterministic part is included in the regression as any prior 
**	detrending has already taken place
*/
proc(3)= wsp(x,d,p);

local T,sum1a,sum1b,s1,zbs,j1,Dxj1,w1,sum2a,sum2b,s2,zfs,j2,Dxj2,w2,As,bs,theta,sum1q,sq1,Dxj1p,Dxlags1,jq1,sum2q,
	sq2,Dxlags2,jq2,Dxj2p,Q,Ainv,estvar,wsstat,k,x1,x2,nb,depf,depb,indf,indb,w1tr,Dx,i1,i2,q1,q2,xksin1,xk,xbs,c;
local    aic,sbc;

if d==0;
c=2;
elseif d==1;
c=3;
endif;

T=rows(x);

sum1a=zeros(p+1,p+1);
sum1b=zeros(p+1,1);


   Dx=trimr(x-lagn(x,1),1,0);
     k       = 0;
     zbs  = trimr(lagn(x,1),1,0);


 if (p > 0) ;
        do until k >= p ;
           k = k + 1 ;
       zbs = zbs~lagn(Dx,k);
        endo ;
     endif ;

     zbs       = trimr(zbs,k,0);


w1=-9999*ones(rows(x),1);

s1=1;
do while s1<=T;

if (1<=s1 and s1<=p+1);   

w1[s1]=0;

elseif  ((p+1)< s1) and (s1<=T-p);

w1[s1]=(s1-p-1)/(T-2*p);

elseif (T-p< s1<=T);

w1[s1]=1;

endif;

s1=s1+1;
endo;


    w1tr  = trimr(w1,k+1,0);
    xbs =trimr(x,k+1,0);

i1=1;
do while i1<=rows(zbs);

sum1a=sum1a+w1tr[i1]*(zbs[i1,.]'*zbs[i1,.]);

sum1b=sum1b+w1tr[i1]*zbs[i1,.]'*xbs[i1];

i1=i1+1;
endo;



w2=trimr(w1,1,p);

sum2a=zeros(p+1,p+1);
sum2b=zeros(p+1,1);

     k       = 0;
  zfs=trimr(x,1,0);

 if (p > 0) ;
zfs=trimr(zfs,0,p);
        do until k >= p ;
           k = k + 1 ;
xksin1=trimr(x,k+1,p-k);
xk=trimr(x,k,p-(k-1));

       zfs = zfs~(-(xksin1-xk));
        endo ;
endif;

     depf     = trimr(x,0,k+1);


i2=1;
do while i2<=rows(zfs);

sum2a=sum2a+(1-w2[i2])*(zfs[i2,.]'*zfs[i2,.]);
sum2b=sum2b+(1-w2[i2])*zfs[i2,.]'*x[i2];

i2=i2+1;
endo;



As=sum1a+sum2a;

bs=sum1b+sum2b;

theta=inv(As)*bs;


sum1q=0;
q1=1;
do while q1<=rows(zbs);

if p>0;

indb=xbs~zbs;

sum1q=sum1q+w1tr[q1]*((1~(-1)*theta')*indb[q1,.]')^2;

elseif p==0;

indb=xbs~zbs[.,1];

sum1q=sum1q+w1tr[q1]*((1~(-1)*theta[1])*indb[q1,.]')^2;

endif;

q1=q1+1;
endo;



sum2q=0;
q2=1;
do while q2<=rows(zfs);


if p>0;

indf=depf~zfs;

sum2q=sum2q+(1-w2[q2])*((1~(-1)*theta[1]~((-1)*theta[2:rows(theta)])')*indf[q2,.]')^2;

elseif p==0;

indf=depf~zfs[.,1];

sum2q=sum2q+(1-w2[q2])*((1~(-1)*theta[1])*indf[q2,.]')^2;

endif;

q2=q2+1;
endo;


Q=sum1q+sum2q;


Ainv=inv(As);

estvar=(Q/(T-p-c))*Ainv[1,1];


wsstat=(theta[1,1]-1)/sqrt(estvar);



aic=ln(Q/T)+2*(p+1)/T;
sbc=ln(Q/T)+(p+1)*ln(T)/T;



retp(wsstat,aic,sbc);
endp;

proc lagn(x,n);
    local y;
    y = shiftr(x', n, (miss(0, 0))');
    retp(y');
endp;


proc detrend(y,d);
local t, u, m, timep, it, iit, xmat, invx, beta, resid, nobs;
     if (d == -1);
         retp(y);
     endif;
     nobs     = rows(y);
     u        = ones(nobs,1);
     if d > 0;
       timep    = zeros(nobs,d);
       t        = seqa(1,1,nobs)/nobs;
       m        = 1;
       do while m <= d;
          timep[.,m] = t^m;
          m = m + 1;
       endo;
       xmat     = u~timep;
     else;
       xmat = u;
     endif;
     invx     = inv(xmat'xmat);
     beta     = invx*(xmat'y);
     resid    = y - xmat*beta;
retp(resid);
endp;




end; 


/** ADF **/
proc(5)=adf(y,maxp,case);/* case=1:Nointercept case=2;intercet, case=3 intercept and trend */
local y_1,dy,c_p,dyp,s,s2,tstar,x,phi,e,se,tstat,aic,sbc,param,resid;
param=zeros(maxp+1,case+maxp)*_dxmiss;
y_1=lag(y);
dy=y-y_1;

c_p=1;dyp=dy;
do while c_p<=maxp;
if c_p==1;dyp=lag(dyp);
else; dyp=dyp~lag(dyp[.,c_p-1]);
endif;
c_p=c_p+1;
endo;
dyp=dyp[maxp+2:rows(dy),.];

aic=zeros(maxp+1,1);
sbc=zeros(maxp+1,1);

c_p=0;
do while c_p<=maxp;

s=dy[maxp+2:rows(dy)];
tstar=rows(s);/**effective # of obs. */

if c_p==0;
    if case==1;x=y_1[maxp+2:rows(dy)];
    elseif case==2;x=y_1[maxp+2:rows(dy)]~ones(tstar,1);
    elseif case==3;x=y_1[maxp+2:rows(dy)]~ones(tstar,1)~seqa(1,1,tstar);
    endif;

phi=s/x;
e=s-x*phi;s2=e'e/(tstar-cols(x));
se=sqrt(diag(s2*invpd(x'x)));
tstat=phi[1]/se[1];
/*rows(dyp)*/
aic[c_p+1]=ln(e'e/rows(dyp))+2*cols(x)/rows(dyp);/** The smaller the better!**/
sbc[c_p+1]=ln(e'e/rows(dyp))+cols(x)*ln(rows(dyp))/rows(dyp);/** The smaller the better!**/

param[c_p+1,1:case+c_p]=phi';
resid=e';

else;
x=x~dyp[.,c_p];
phi=s/x;
e=s-x*phi;s2=e'e/(rows(dyp)-cols(x));
se=sqrt(diag(s2*invpd(x'x)));
tstat=tstat|(phi[1]/se[1]);

aic[c_p+1]=ln(e'e/rows(dyp))+2*cols(x)/rows(dyp);/** The smaller the better!**/
sbc[c_p+1]=ln(e'e/rows(dyp))+cols(x)*ln(rows(dyp))/rows(dyp);/** The smaller the better!**/

param[c_p+1,1:case+c_p]=phi';
resid=resid|e';


endif;

c_p=c_p+1;
endo;



retp(tstat,aic,sbc,param,resid);
endp;


/** case=2: with intercept */
/** case=3: with intercept and trend **/
/** widchoice (window size)
        1: (T^(1/3))*0.75
        2: 2*sqrt(T)
modified the code by J. Lee, Sep. 1990*/ 

proc(1)=kpss(y,case,widchoice);

local dy,isw,BETA,n,n1,tt,x,e,s,top,sum1,width1,width,it,is,sum2,wsl,bot,stat;

dy=y;
isw = case-2;
   /* isw = 0  without trend , isw = 1  with trend */

  n = rows(dy);
 n1 = n - 1 ;

 tt = seqa(1,1,n);
 if isw == 0;    x = ones(n,1) ;
 elseif isw == 1;    x = ones(n,1) ~ tt;  
endif;
 beta = inv(x'x)*x'dy;
    e = dy - x * beta ;
    s = cumsumc(e);

    top = s's / (n^2) ;
    sum1 = e'e / n;

 /*   lagloop=4;*/
if widchoice==1;
width=round((n^(1/3))*0.75);
elseif widchoice==2;
width=round((n^(1/2))*2);
elseif widchoice==3;
width=round((n^(1/3))*0.75+(((n^(1/2))*2-(n^(1/3))*0.75)/2));
endif;
/*    do until lag == lagloop;*/

      width1=width+1;
      it=1; is=1;
      sum2=0;
      do while is<width+1;
        it=is+1;
        wsl=1-is/width1;
        do while it< n+1;
          sum2=sum2+wsl.*e[it,1].*e[it-is,1];
          it=it+1;
        endo;
        is=is+1;
      endo;

      sum2=2*sum2/n;

      bot = sum1 + sum2 ;
      stat = top / bot ;

/*      lag = lag + 1;

    endo;*/

retp(stat);
endp;



/*** get starting row **/
proc(1)=getst(y);
local i;

i=1;
do while i<=rows(y);
if y[i].==_dxmiss;i=i+1;
else; goto nomore3;
endif;
endo;
nomore3:
return(i);
endp;